from enum import Enum


class Alertlevel(Enum):
    """
    Enum that represents the pager alert levels of earthquake events.
    """
    GREEN = "green"
    """
    Green pager level
    """

    YELLO = "yello"
    """
    Yellow pager level
    """

    ORANGE = "orange"
    """
    Orange pager level
    """

    RED = "red"
    """
    Red pager level
    """



from datetime import datetime


class SingleResult:
    """
    This class represents the result data of a single earthquake event. In the USGS REST API,
    searching for a range (time/location) gives a collection of events, which is named as
    **FeatureCollection** in GeoJSON standard. Searching with a particular event ID gives us a
    single event, which is named **Feature** in GeoJSON standard.

    This class accepts all the json data that the query generated, processes the data, and provides
    users all the necessary APIs for getting ordered, sorted and simplified results. The APIs are specially
    designed to include all functionality of the USGS API.

    Example:
    ::
        event_id = "usc000lvb5"
        single_result = EarthquakeQuery.search_by_event_id(event_id=event_id)
        single_result.get_raw_json()


    """
    def __init__(self, result_json):
        """
		Constructor:
			Saves the query result json string as class attribute

		:param result_json: str, query result json string
		"""

        self.json_raw = result_json

    def get_raw_json(self) -> str:
        """
        Get a raw json file which describes a single earthquake

        :return: str, raw json string
        """
        return self.json_raw

    def get_raw_properties(self) -> dict:
        """
        Get properties of a individual earthquake, i.e. mag, title, etc.

        :return: dict, a dictionary contains all properties attributes of an earthquake
        """
        return self.json_raw["properties"]

    def get_raw_geometry(self) -> dict:
        """
        Get geometry of an individual earthquake, i.e. type and coordinates

        :return: dict, a dictionary, contains all geometry info of an earthquake
        """
        return self.json_raw["geometry"]

    def get_coordinates(self) -> list:
        """
        Get coordinates of where the individual earthquake source centered

        :return: list, a list of the coordinates, which has latitude, longitude, and elevation info
        """
        return self.json_raw["geometry"]["coordinates"]

    def get_magnitude(self) -> float:
        """
        Get the magnitude of the earthquake, measuring of the size of the earthquake source

        :return: float, the magnitude of the earthquake
        """
        return self.json_raw["properties"]["mag"]

    def get_title(self) -> str:
        """
        Get the title of the earthquake, containing the basic info describing the earthquake,
        i.e. "M 6.5 - 32km W of Sola, Vanuatu"

        :return: str, the title of the earthquake
        """
        return self.json_raw["properties"]["title"]

    def get_location_string(self) -> str:
        """
        Get the location of the earthquake, describing the direction from the nearby city.
        For example: '32km W of Sola, Vanuatu'

        :return: str, the location of the earthquake
        """
        return self.json_raw["properties"]["place"]

    def get_epoch_time(self) -> int:
        """
        Get the epoch time of the earthquake happened

        :return: int, the epoch time
        """
        return self.json_raw["properties"]["time"]

    def get_datetime(self) -> datetime:
        """
        Convert the epoch time to datetime in Python

        :return: datetime, the converted datetime object from epoch time
        """
        return datetime.fromtimestamp(self.json_raw["properties"]["time"] / 1000.0)

    def get_webpage_url(self) -> str:
        """
        Get the web page URL of this particular earthquake event

        :return: str, the web page URL of this event
        """
        return self.json_raw["properties"]["url"]



from .earthquake_query import EarthquakeQuery
from .location import Rectangle, Circle, GeoRectangle, GeoCircle
from .result_collection import ResultCollection
from .single_result import SingleResult
from .timeframe import TimeFrame
from .enum.alertlevel import Alertlevel
from .enum.catalog import Catalog
from .enum.contributor import Contributor
from .enum.delete import Delete
from .enum.magnitude import Magnitude
from .enum.nodata import Nodata
from .enum.origin import Origin
from .enum.supersede import Supersede



import os
from enum import Enum

import requests
import urllib.parse
from .address import _Address


class IncludeNeighborhood(Enum):
    """
    Enum values of include_neighborhood_flag, which indicates whether to include
    the neighborhood information when acquiring geolocation of an address.

    INCLUDE_NEIGHBORHOOD is include neighborhood information when available.
    NOT_INCLUDE_NEIGHBORHOOD is not include neighborhood information.
    """
    INCLUDE_NEIGHBORHOOD = 1
    NOT_INCLUDE_NEIGHBORHOOD = 0


class IncludeAdditionalResponse(Enum):
    """
    Enum values of addition_response_flag, which indicates different response format.

    INCLUDE_QUERY_PARSE is to specify that the response will be parsed into address values,
    such as addressLine, locality, adminDistrict, and postalCode.
    INCLUDE_CISO2 is to include the two-letter ISO country code.
    INCLUDE_BOTH is to include both INCLUDE_QUERY_PARSE and INCLUDE_CISO2 functionalities.
    """
    INCLUDE_QUERY_PARSE = "queryParse"
    INCLUDE_CISO2 = "ciso2"
    INCLUDE_BOTH = "queryParse,ciso2"


class _GeoLocationInfo:
    """
    A geolocation instance, containing three important information, including the center coordinates,
    boundary box and the address of a user specified address.

    This class will be internally used by the GeocodeQuery, GeoRectangle and GeoCircle classes. Users
    do not need to worry about the initialization process.
    """

    def __init__(self, center, bbox, address):
        """
        Construct a GeoLocationInfo class.

        :param center: center coordinates of a specified location
        :param bbox: bounding box of a specified location
        :param address: formatted address of a specified location
        """
        self.center = center
        self.address = address
        self.bounding_box = bbox

    def get_address(self):
        """
        Get the formatted address of a specified location.

        :return address of the location.
        """
        return self.address

    def get_center(self):
        """
        Get the center coordinates of a specified location.

        :return center coordinates of the location.
        """
        return self.center

    def get_bounding_box(self):
        """
        Get the bounding box of a specified location.

        :return bounding box of the location.
        """
        return self.bounding_box


class _GeocodeQuery:
    """
    An instance of the GeocodeQuery. This class is aimed to get the geological information
    of the user specified address.

    This class will be internally used by the GeoRectangle and GeoCircle class. User do not need to
    worry about the query sending and result parsing processes.

    This class is not thread safe.
    """

    query_url = "http://dev.virtualearth.net/REST/v1/Locations"

    def __init__(self, key_path):
        """
        Construct a GeocodeQuery instance.
        The user specified key will be read automatically.
        """
        self.key = self.read_key(key_path)
        self.query = None
        self.include_neighborhood_flag = None
        self.max_results = None
        self.include_additional_response_flag = None

    def get_key(self):
        """
        Get the user specified key.

        :return user key.
        """
        return self.key

    def read_key(self, key_path):
        """
        Read the user specified key.

        :raise KeyError - key file doesn't exist
                          cannot read user key from key file.
        """
        if os.path.exists(key_path):
            with open(key_path, 'r') as file:
                try:
                    if os.path.getsize(key_path) != 0:
                        key = file.read().replace('\n', '')
                        return key
                    else:
                        raise KeyError("Please put user key in " + key_path)
                except:
                    raise KeyError("Cannot Read Default Key")
        else:
            raise KeyError("Cannot read key file")

    def set_include_neighborhood(self, include_neighborhood_flag: IncludeNeighborhood) -> '_GeocodeQuery':
        """
        Set whether the GeocodeQuery should include the neighborhood information if available.

        :param include_neighborhood_flag: if include_neighborhood_flag is INCLUDE_NEIGHBORHOOD, include neighborhood
                                          information if available.
                                          otherwise, not include neighborhood information.
        :type include_neighborhood_flag: IncludeNeighborhood
        :return GeocodeQuery
        """
        self.include_neighborhood_flag = include_neighborhood_flag
        return self

    def get_include_neighborhood(self):
        """
        Get the status of GeocodeQuery about whether to include the neighborhood information if available.

        :return status of whether to include the neighborhood information if available.
        """
        return self.include_neighborhood_flag

    def set_max_results(self, max_results) -> '_GeocodeQuery':
        """
        Set the max result of one time GeocodeQuery.

        max_results should be in the range of 1 (inclusive) and 20 (inclusive).

        :raise TypeError: if the input is not an integer.
               ValueError: if the input is out of bounds.
        :param max_results: user specified max result number for one time GeocodeQuery.
        :type max_results: int
        :return: GeocodeQuery
        """
        if isinstance(max_results, int):
            if 1 <= max_results <= 20:
                self.max_results = max_results
                return self
            else:
                raise ValueError("set_max_results input out of bounds")
        else:
            raise TypeError("set_max_results input is not an integer ")

    def get_max_results(self):
        """
        Get the max results of the GeocodeQuery.

        :return max results.
        """
        return self.max_results

    def set_include_additional_response(self, addition_response_flag: IncludeAdditionalResponse) -> '_GeocodeQuery':
        """
        Set whether the GeocodeQuery should include additional information in the response.

        :param addition_response_flag: if addition_response_flag is set to INCLUDE_QUERY_PARSE, include the address
                                        values in the response.
                                        if addition_response_flag is set to INCLUDE_CISO2, include the two-letter ISO
                                        country code in the response.
                                        otherwise, include both of INCLUDE_QUERY_PARSE and INCLUDE_CISO2.
        :type addition_response_flag: IncludeAdditionalResponse
        :return GeocodeQuery
        """
        self.include_additional_response_flag = addition_response_flag
        return self

    def get_include_additional_response(self):
        """
        Get the status of GeocodeQuery about whether including additional information in the response.

        :return status of whether including additional information in the response.
        """
        return self.include_additional_response_flag

    def set_one_line_Address(self, one_line_address):
        """
        Set the address of the GeocodeQuery

        :param one_line_address: query address
        """
        self.query = urllib.parse.quote(one_line_address)

    def get_one_line_Address(self):
        """
        Get the address of the GeocodeQuery.

        :return address to be queried.
        """
        return self.query

    def send_query(self):
        """
        Send the geocode query to the server and get the result in json format.

        :return response in json format.
        """
        self.query_url += "/" + self.get_one_line_Address()
        request_dict = {}
        if self.get_include_additional_response() is not None:
            request_dict["include"] = self.get_include_additional_response().value
        if self.get_include_neighborhood() is not None:
            request_dict['includeNeighborhood'] = self.get_include_neighborhood().value
        if self.get_max_results() is not None:
            request_dict['maxResults'] = self.get_max_results()
        request_dict['key'] = self.get_key()
        r = requests.get(self.query_url, request_dict)
        return r.json()

    def get_address_boxing(self, response):
        """
        Get the geolocation information of the specified

        :param response: response get from the send operation.
        :return GeoLocationInfo of a specified address
        """
        resourceSets = response['resourceSets']
        geo_location = None
        if len(resourceSets) > 0:
            resource_item = response['resourceSets'][0]
            if len(resource_item) > 0 and len(resource_item['resources']) > 0:
                resources = resource_item['resources']
                single_source_item = resources[0]
                single_source_item_center = single_source_item['point']['coordinates']
                single_source_item_bounding_box = single_source_item['bbox']
                address = single_source_item['address']
                addressLine = self._assign_key('addressLine', address)
                adminDistrict = self._assign_key('adminDistrict', address)
                adminDistrict2 = self._assign_key('adminDistrict2', address)
                countryRegion = self._assign_key('countryRegion', address)
                countryRegionIso2 = self._assign_key('countryRegionIso2', address)
                formattedAddress = self._assign_key('formattedAddress', address)
                locality = self._assign_key('locality', address)
                postalCode = self._assign_key('postalCode', address)
                single_address = _Address().set_address_line(addressLine).set_admin_district(adminDistrict) \
                    .set_admin_district2(adminDistrict2).set_country_region(countryRegion) \
                    .set_country_region_iso2(countryRegionIso2).set_formatted_address(formattedAddress) \
                    .set_locality(locality).set_postal_code(postalCode)
                geo_location = _GeoLocationInfo(single_source_item_center, single_source_item_bounding_box,
                                                single_address)
        return geo_location

    def _assign_key(self, key, address_dict):
        if key in address_dict.keys():
            return address_dict[key]
        else:
            return ""



from enum import Enum


class Contributor(Enum):
    """
    A Enum that represents all earthquake data contributors. Use this enum in EarthquakeQuery class to set its
    parameters related to contributors.
    """
    CONTRIBUTOR_ADMIN = "admin"
    """
    Admin - Administrator
    """

    CONTRIBUTOR_AK = "ak"
    """
    AK - Alaska Earthquake Center
    """

    CONTRIBUTOR_AT = "at"
    """
    AT - National Tsunami Warning Center
    """

    CONTRIBUTOR_ATLAS = "atlas"
    """
    ATLAS - ShakeMap Atlas
    """

    CONTRIBUTOR_AV = "av"
    """
    AV - Alaska Volcano Observatory
    """

    CONTRIBUTOR_CGS = "cgs"
    """
    California Geological Survey
    """

    CONTRIBUTOR_CI = "ci"
    """
    CI - California Integrated Seismic Network: Southern, California Seismic Network (Caltech/USGS Pasadena and Partners) 
    and Southern California Earthquake Data Center
    """

    CONTRIBUTOR_EW = "ew"
    """
    EW - EW database
    """

    CONTRIBUTOR_HV = "hv"
    """
    HV - Hawaii Volcano Observatory
    """

    CONTRIBUTOR_ISMP = "ismp"
    """
    ISMP - USGS Induced Seismicity Project
    """

    CONTRIBUTOR_LD = "ld"
    """
    LD - Lamont-Doherty Cooperative Seismographic Network
    """

    CONTRIBUTOR_MB = "mb"
    """
    MB - Montana Bureau of Mines and Geology
    """

    CONTRIBUTOR_NC = "nc"
    """
    NC - California Integrated Seismic Network: Northern California Seismic System 
    (UC Berkeley, USGS Menlo Park, and Partners)
    """

    CONTRIBUTOR_NM = "nm"
    """
    NM - New Madrid Seismic Network
    """

    CONTRIBUTOR_NN = "nn"
    """
    NN - Nevada Seismological Laboratory
    """

    CONTRIBUTOR_NP = "np"
    """
    NP - NP Contributor
    """

    CONTRIBUTOR_OFFICIAL = "official"
    """
    Official - United States Geological Survey
    """

    CONTRIBUTOR_OK = "ok"
    """
    OK - Oklahoma Geological Survey Statewide Network, Norman, USA (aka TUL, OGS)
    """

    CONTRIBUTOR_PR = "pr"
    """
    PR - Puerto Rico Seismic Network
    """

    CONTRIBUTOR_PT = "pt"
    """
    PT - Pacific Tsunami Warning Center
    """

    CONTRIBUTOR_SE = "se"
    """
    SE - Center for Earthquake Research and Information
    """

    CONTRIBUTOR_TX = "tx"
    """
    TX - Texas Seismological Network (TexNet) (aka TEXNET, BEG)
    """

    CONTRIBUTOR_US = "us"
    """
    US - USGS National Earthquake Information Center, PDE (aka GS, NEIC)
    """

    CONTRIBUTOR_UU = "uu"
    """
    UU - University of Utah Seismograph Stations
    """

    CONTRIBUTOR_UW = "uw"
    """
    UW - Pacific Northwest Seismic Network
    """


from enum import Enum


class Supersede(Enum):
    """
    Enum values of includesuperseded flag in the query parameters
    """
    INCLUDE_SUPERSEDED = True
    """
    Include superseded events in the query result
    """

    NOT_INCLUDE_SUPERSEDED = False
    """
    Not include superseded events in the query result
    """



class ResultCollection:
	"""
	This is a class that represents an collection of earthquake events. An query can be a single event
	search or a ranged event search (i.e. difference between searching for a particular ID and a range
	of magnitudes). When the query specified a range of constraints, it will return a collection of
	earthquake events.

	This class accepts all the json data that the query generated, processes the data, and provides
	users all the necessary APIs for getting ordered, sorted and simplified results. The APIs are specially
	designed to include all functionality of the USGS API.

	Ordering types:
		The ResultCollection class supports multiple ordering types, all of them are represented in Strings

		- time: timestamp of the earthquake event
		- mag: short for magnitude, the magnitude of the earthquake
		- latitude: latitude of the epicenter
		- longitude: longitude of the epicenter
		- depth: depth of the epicenter

	Sample usage:
	::
		earthquake_query = EarthquakeQuery(time=None, Location=[rectangle, circle, geo_rectangle, geo_circle])
		result = earthquake_query.search()
		data1 = result.get_all_simplified_data(order_by="time")
		data2 = result.get_all_magnitudes(order_by="mag")

	"""
	def __init__(self, result_json_list):
		"""
		Constructor:
			Initialize the result object, remove all duplicating earthquakes when initializing

		:param result_json_list: list, the original list of json strings returned by all the requests
		                         made in the query
		"""
		self.json_raw = result_json_list
		self.json_combined = self._combine_json_list(result_json_list)

	def _combine_json_list(self, json_list: list):
		unique_results = self._combine_unique_results(json_list)
		combined_metadata = self._combine_metadata(json_list, len(unique_results))
		combined_bbox = self._combine_boundary_box(json_list)

		base_json = {"type": json_list[0]["type"],
					 "metadata": combined_metadata,
					 "features": unique_results,
					 "bbox": combined_bbox}

		return base_json

	def _combine_metadata(self, json_list: list, count: int):
		base_metadata = {"generated": [],
						 "url": [],
						 "title": json_list[0]["metadata"]["title"],
						 "status": 200,
						 "api": json_list[0]["metadata"]["api"],
						 "count": count}

		for data in json_list:
			base_metadata["generated"].append(data["metadata"]["generated"])
			base_metadata["url"].append(data["metadata"]["url"])

		return base_metadata

	def _combine_unique_results(self, json_list):
		result_id_set = set()
		unique_results = []

		for data in json_list:
			for result in data["features"]:
				ids = result["properties"]["ids"].strip(",").split(",")
				if any(x in result_id_set for x in ids):
					continue
				else:
					result_id_set.update(ids)
					unique_results.append(result)
		return unique_results

	def _combine_boundary_box(self, json_list):
		base_bbox = []
		for data in json_list:
			try:
				base_bbox.append(data["bbox"])
			except:
				continue
		return base_bbox

	def _get_sorted_results(self, order_by, descending):
		if order_by == "latitude":
			return sorted(self.json_combined["features"], key=lambda i: i["geometry"]["coordinates"][1], reverse=descending)
		if order_by == "longitude":
			return sorted(self.json_combined["features"], key=lambda i: i["geometry"]["coordinates"][0], reverse=descending)
		if order_by == "depth":
			return sorted(self.json_combined["features"], key=lambda i: i["geometry"]["coordinates"][2], reverse=descending)

		return sorted(self.json_combined["features"], key=lambda i: i["properties"][order_by], reverse=descending)

	def get_combined_json(self) -> dict:
		"""
		Get the combined raw json dict of the collection

		:return: dict, the raw json dict of the query
		"""
		return self.json_combined

	def get_metadata(self) -> dict:
		"""
		Get the combined metadata of the result collection

		:return: dict, a dictionary containing all metadata values
		"""
		return self.json_combined["metadata"]

	def get_boundary_box(self) -> list:
		"""
		Get the combined boundary box of the result collection.
		Combined boundary box means taking all max and min lat/long, and pick the extreme values.
		The updated combined boundary box should contain all the query regions.

		:return: list of coordinates, specifying the boundary box
		"""
		return self.json_combined["bbox"]

	def get_data_by_keys(self, keys:list, order_by="time", descending=True) -> dict:
		"""
		Get a simplified version of the results by specifying which keys the users wish to get.
		For example, if the user wants only the title, magnitude and time, he should pass the list
		["title", "mag", "time"] to this function.

		:param keys: list, list of keys. Each key should be present in the GeoJSON
		:param order_by: str, ordering type
		:param descending: bool, indicates whether the ordered data should be in descending order
		:return: dict, key value pairs of the result earthquake data
		"""
		to_return = []
		results = self._get_sorted_results(order_by, descending)

		for result in results:
			d = dict.fromkeys(keys)
			for key in keys:
				if key in result["properties"]:
					d[key] = result["properties"][key]
				if key in result["geometry"]:
					d[key] = result["geometry"][key]
				if key == "id":
					d[key] = result["id"]
			to_return.append(d)

		return to_return

	def get_all_earthquake_data(self, order_by="time", descending=True) -> list:
		"""
		Get all earthquakes data from the returned query

		:param order_by: str, ordering mode
		:param descending: bool, indicates whether it is in descending order
		:return: list, list of earthquakes in dict form
		"""
		return self._get_sorted_results(order_by, descending)

	def get_all_details_url(self, order_by="time", descending=True) -> list:
		"""
		Get all earthquake detailed data urls from the returned query

		:param order_by: str, ordering mode
		:param descending: bool, indicates whether it is in descending order
		:return: list of dicts, list of urls for detailed earthquake information
		"""
		to_return = []
		results = self._get_sorted_results(order_by, descending)

		for result in results:
			to_return.append({"title": result["properties"]["title"],
							  "ids": result["properties"]["ids"],
							  "detail": result["properties"]["detail"]})
		return to_return

	def get_all_simplified_data(self, order_by="time", descending=True) -> list:
		"""
		Get all earthquake data in a simplified term, keys will be title, coordinates, magnitude and depth

		:param order_by: str, ordering mode
		:param descending: bool, indicates whether it is in descending order
		:return: list, list of dicts representing the earthquakes
		"""
		to_return = []
		results = self._get_sorted_results(order_by, descending)

		for result in results:
			to_return.append({"title": result["properties"]["title"],
							  "mag": result["properties"]["mag"],
							  "time": result["properties"]["time"],
							  "coordinates": result["geometry"]["coordinates"]})
		return to_return

	def get_all_magnitudes(self, order_by="mag", descending=True) -> list:
		"""
		Get a list of earthquake magnitudes

		:param order_by: str, ordering mode
		:param descending: bool, indicates whether it is in descending order
		:return: list, list of earthquake magnitudes
		"""
		ordered = self._get_sorted_results(order_by, descending)
		return [x["properties"]["mag"] for x in ordered]

	def get_all_coordinates(self, order_by="time", descending=True) -> list:
		"""
		Get a list of earthquake coordinates

		:param order_by: str, ordering mode
		:param descending: bool, indicates whether it is in descending order
		:return: list, list of tuples representing coordinates
		"""
		ordered = self._get_sorted_results(order_by, descending)
		return [x["geometry"]["coordinates"][:2] for x in ordered]

	def get_all_3d_coordinates(self, order_by="time", descending=True) -> list:
		"""
		Get a list of earthquake coordinates, with the third dimension representing depth

		:param order_by: str, ordering mode
		:param descending: bool, indicates whether it is in descending order
		:return: list, list of 3-tuples representing coordinates
		"""
		ordered = self._get_sorted_results(order_by, descending)
		return [x["geometry"]["coordinates"] for x in ordered]

	def get_all_depths(self, order_by="time", descending=True) -> list:
		"""
		Get a list of earthquake depths

		:param order_by: str, ordering mode
		:param descending: bool, indicates whether it is in descending order
		:return: list, list of int representing depths
		"""
		ordered = self._get_sorted_results(order_by, descending)
		return [x["geometry"]["coordinates"][2] for x in ordered]
	
	def get_all_titles(self, order_by="time", descending=True) -> list:
		"""
		Get a list of earthquake event titles

		:param order_by: str, ordering mode
		:param descending: bool, indicates whether it is in descending order
		:return: list, list of titles
		"""
		ordered = self._get_sorted_results(order_by, descending)
		return [x["properties"]["title"] for x in ordered]

	def get_number_of_earthquakes(self) -> int:
		"""
		Get the total number of earthquakes in the query result

		:return: int, integer presenting the total number of earthquakes
		"""
		return len(self.json_combined["features"])



from enum import Enum


class Origin(Enum):
    """
    Enum that represents the includeallorigins flag of earthquake query parameter.
    """
    INCLUDE_ALL_ORIGINS = True
    """
    Include all origins in the query result
    """

    NOT_INCLUDE_ALL_ORIGINS = False
    """
    Not include all origins in the query result
    """



import datetime


class TimeFrame:
    """
    Clients can use this class to specify the time frame for a query.
    Clients should use the python built-in datetime object to initialize the time.
    Example:
    ::
        start_time = datetime.now()- timedelta(20)
        end_time = datetime.now()
        timeframe = TimeFrame(start_time, end_time)

    After initializing, build a list TimeFrame objects and put it in the EarthquakeQuery constructor.
    Example:
    ::
        start_time = datetime.now() - timedelta(20)
        end_time = datetime.now()
        timeframe1 = TimeFrame(start_time, end_time)
        timeframe2 = TimeFrame(datetime(2020, 1, 1), datetime(2021, 1, 1))
        earthquake_query = EarthquakeQuery(time=[timeframe1, timeframe2])
        result = earthquake_query.search()
    """

    def __init__(self, start_time, end_time, update_after=None):
        """
        Create a TimeFrame object with start_time, end_time and update_after.

        :param start_time: The start time of the timeframe. Limit to events on or after the specified start time.
        :param end_time: The end time of the timeframe. Limit to events on or before the specified end time.
        :param update_after: The update after time. Limit to events updated after the specified time.
        :type start_time: datetime.datetime
        :type end_time: datetime.datetime
        :type update_after: datetime.datetime
        :raises TypeError: when any parameter is not an instance of datetime.datetime
        :raises ValueError: when start_time is later than end_time
        """
        if not isinstance(start_time, datetime.datetime):
            raise TypeError("start_time should be instance of datetime.datetime")
        if start_time > end_time:
            raise ValueError("start_time should not be later than end_time")

        if not isinstance(end_time, datetime.datetime):
            raise TypeError("end_time should be instance of datetime.datetime")
        if start_time > end_time:
            raise ValueError("start_time should not be later than end_time")

        if update_after is not None and not isinstance(update_after, datetime.datetime):
            raise TypeError("update_after should be instance of datetime.datetime")

        self.end_time = end_time
        self.start_time = start_time
        self.update_after = update_after

    def set_start_time(self, start_time) -> 'TimeFrame':
        """
        Limit to events on or after the specified start time. Default value: NOW - 30 days

        :param start_time: The start time. Using the datetime object.
        :type start_time: datetime.datetime
        :raises TypeError: when start_time is not an instance of datetime.datetime
        :raises ValueError: when start_time is later than end_time
        :return: the updated timeframe object.
        """
        if not isinstance(start_time, datetime.datetime):
            raise TypeError("start_time should be instance of datetime.datetime")
        if start_time > self.end_time:
            raise ValueError("start_time should not be later than end_time")
        self.start_time = start_time
        return self

    def set_end_time(self, end_time) -> 'TimeFrame':
        """
        Limit to events on or before the specified end time. Default value: present time

        :param end_time: The end time. Using the datetime object.
        :type end_time: datetime.datetime
        :raises TypeError: when end_time is not an instance of datetime.datetime
        :raises ValueError: when start_time is later than end_time
        :return: the updated timeframe object.
        """
        if not isinstance(end_time, datetime.datetime):
            raise TypeError("end_time should be instance of datetime.datetime")
        if self.start_time > end_time:
            raise ValueError("start_time should not be later than end_time")
        self.end_time = end_time
        return self

    def set_update_after(self, update_after) -> 'TimeFrame':
        """
        Limit to events updated after the specified time. Default value: None

        :param update_after: The update after time. Using the datetime object.
        :type update_after: datetime.datetime
        :raises TypeError: when update_after is not an instance of datetime.datetime
        :return: the updated timeframe object.
        """
        if not isinstance(update_after, datetime.datetime):
            raise TypeError("update_after should be instance of datetime.datetime")
        self.update_after = update_after
        return self

    def get_start_time_string(self) -> str:
        """
        Get the start_time string in ISO8601 Date/Time format.

        :return: str, the start_time string
        """
        return self.start_time.isoformat().split(".")[0]

    def get_end_time_string(self) -> str:
        """
        Get the end_time string in ISO8601 Date/Time format.

        :return: str, he end_time string
        """
        return self.end_time.isoformat().split(".")[0]

    def get_update_after_string(self) -> str:
        """
        Get the update_after string in ISO8601 Date/Time format.
        Please use the function is_update_after_set() to make sure update_after is set before calling this method,
        otherwise it will raise NullPointerException.

        Example:
        ::
            time_frame = TimeFrame()
            if time_frame.s_update_after_set():
                time_frame.get_update_after_string()

        :return: str, the update_after string
        """
        return self.update_after.isoformat().split(".")[0]

    def is_update_after_set(self) -> bool:
        """
        Check if the update_after time is set.

        :return: bool, true if update_after is set.
        """
        return self.update_after is not None



import requests
import urllib.parse
from typing import List

from .timeframe import TimeFrame
from .location import Location
from .enum.catalog import Catalog
from .enum.contributor import Contributor
from .enum.magnitude import Magnitude
from .enum.origin import Origin
from .enum.alertlevel import Alertlevel
from .enum.delete import Delete
from .enum.supersede import Supersede
from .result_collection import ResultCollection
from .single_result import SingleResult
from .key import _Key


class EarthquakeQuery:
    """
    A base class handles setting the parameters, sending the request to the USGS Earthquake Catalog API and getting the
    query results.

    The parameters can be set in two ways.
    1. The constructors have a time parameter to set a list of **TimeFrame**, a location parameter to set a list of
    **Location**, and all other parameters can be set using their parameter names.
    For example,
    ::
        query = EarthquakeQuery(time=[TimeFrame(datetime(2010, 1, 1), datetime(2015, 1, 1))], maxdepth=1.0)

    2. Use the corresponding set method to set the parameter value.
    For example,
    ::
        query = EarthquakeQuery()
        query.set_time([TimeFrame(datetime(2010, 1, 1), datetime(2015, 1, 1))])
        query.set_max_depth(1.0)

    Two types of result cam be returned:
        1. The client can use the constructor, different set methods to set the parameters, and then call search() to get
        the result wrapped by an instance of ResultCollection. For how to use the ResultCollection, please see refer to
        the ResultCollection class.

        2. The client can call the search_by_event_id() method to get detailed information of a single earthquake by giving
        its event id wrapped by an instance of SingleResult. For how to use the SingleResult, please see refer to
        the SingleResult class.

    Note:
        If multiple time frames and locations are set, the search() method actually does multiple search to get multiple
        results, then return the combined result in a ResultCollection instance.

    Example:
        A typical usage to search a collection of results is:
        ::
            query = EarthquakeQuery()
            query.set_time([TimeFrame(datetime(2010, 1, 1), datetime(2015, 1, 1))])
            query.set_max_depth(1.0)
            result = query.search()

        A typical usage to search a specific earthquake by its event id is:
        ::
            event_id = "usc000lvb5"
            result = EarthquakeQuery.search_by_event_id(event_id)

    For detail of all available parameter, Please refer to https://earthquake.usgs.gov/fdsnws/event/1/,
    or see the docs of different set methods for each parameter
    """

    _attribute_map = {
        "catalog": "_query_catalog",
        "contributor": "_query_contributor",
        "includeallmagnitudes": "_query_include_all_magnitudes_flag",
        "includeallorigins": "_query_include_all_origins_flag",
        "includedeleted": "_query_include_all_deleted_flag",
        "includesuperseded": "_query_include_all_superseded_flag",
        "limit": "_query_limit",
        "maxdepth": "_query_max_depth",
        "maxmagnitude": "_query_max_magnitude",
        "mindepth": "_query_min_depth",
        "minmagnitude": "_query_min_magnitude",
        "alertlevel": "_query_alert_level",
        "eventtype": "_query_event_type",
        "maxcdi": "_query_max_cdi",
        "maxgap": "_query_max_gap",
        "maxmmi": "_query_max_mmi",
        "maxsig": "_query_max_significance",
        "mincdi": "_query_min_cdi",
        "minfelt": "_query_min_feltcount",
        "mingap": "_query_min_gap",
        "minsig": "_query_min_significance",
        "nodata": "_query_nodata",
        "producttype": "_query_product_type",
        "productcode": "_query_product_code",
        "reviewstatus": "_query_review_status"
    }

    _key_to_set_method_map = {
        "catalog": "set_catalog",
        "contributor": "set_contributor",
        "includeallmagnitudes": "set_include_all_magnitudes",
        "includeallorigins": "set_include_all_origins",
        "includedeleted": "set_include_deleted",
        "includesuperseded": "set_include_superseded",
        "limit": "set_limit",
        "maxdepth": "set_max_depth",
        "maxmagnitude": "set_max_magnitude",
        "mindepth": "set_min_depth",
        "minmagnitude": "set_min_magnitude",
        "alertlevel": "set_alertlevel",
        "eventtype": "set_eventtype",
        "maxcdi": "set_max_cdi",
        "maxgap": "set_max_gap",
        "maxmmi": "set_max_mmi",
        "maxsig": "set_max_significance",
        "mincdi": "set_min_cdi",
        "minfelt": "set_min_feltcount",
        "mingap": "set_min_gap",
        "minsig": "set_min_significance",
        "nodata": "set_nodata_code",
        "producttype": ("set_product_limit", 0),
        "productcode": ("set_product_limit", 1),
        "reviewstatus": "set_reviewstatus"
    }

    _base_url = "https://earthquake.usgs.gov/fdsnws/event/1/query"

    def __init__(self, time: List[TimeFrame] = None, location: List[Location] = None, **kwargs):
        """
        Constructor:
            Construct a Query instance. kwargs include parameters of Other and Extension

        :param time: a list of TimeFrame instances
        :type time: List[TimeFrame]
        :param location: a list of Location instances
        :type location: List[Location]
        :param kwargs: parameters of Other and Extension
        :raises ValueError:  If key is not a parameter name
        """
        for key in kwargs.keys():
            if key not in EarthquakeQuery._key_to_set_method_map:
                raise ValueError(key + " is not a parameter")

        self._query_catalog = None
        self._query_contributor = None
        self._query_eventid = None
        self._query_include_all_magnitudes_flag = None
        self._query_include_all_origins_flag = None
        self._query_include_all_deleted_flag = None
        self._query_include_all_superseded_flag = None
        self._query_limit = 20000
        self._query_max_depth = None
        self._query_min_depth = None
        self._query_max_magnitude = None
        self._query_min_magnitude = None

        self._query_alert_level = None
        self._query_event_type = None
        self._query_max_cdi = None
        self._query_max_gap = None
        self._query_max_mmi = None
        self._query_max_significance = None
        self._query_min_cdi = None
        self._query_min_feltcount = None
        self._query_min_gap = None
        self._query_min_significance = None
        self._query_nodata = None
        self._query_product_type = None
        self._query_product_code = None
        self._query_review_status = None

        self._query_time: List[TimeFrame] = [None] if time is None else time
        self._query_location: List[Location] = [None] if location is None else location

        # set kwagrs
        method_param_map = {}
        for key, value in kwargs.items():
            # get the set method name that needs to be called
            dest = EarthquakeQuery._key_to_set_method_map[key]
            # if the method only needs one parameter
            if type(dest) == str:
                # get the reference to the method
                method_to_call = getattr(self, dest)
                # call the method
                method_to_call(value)
            else:
                # if the method needs two parameters
                # get the method name
                dest_method_name = dest[0]
                # get the position for the key value
                position = dest[1]
                # if this is the first time that one of two parameters are set
                if dest_method_name not in method_param_map:
                    # save the first parameter
                    method_param_map[dest_method_name] = {position: value}
                else:
                    # if the first parameter has been saved
                    method_param_map[dest_method_name][position] = value
                    # build two parameters in a list
                    param_list = [method_param_map[dest_method_name][0], method_param_map[dest_method_name][1]]
                    # get the method reference
                    method_to_call = getattr(self, dest_method_name)
                    # call the method
                    method_to_call(*param_list)

    @staticmethod
    def set_geocode_key_path(key_path: str):
        """
        Set the geocode key file path.
        :param key_path: the path of the key file
        :raises ValueError:  If the key path is not a string
        """
        if not isinstance(key_path, str):
            raise TypeError("key_path should be a string")

        _Key.key_path = key_path

    @staticmethod
    def search_by_event_id(event_id: str) -> SingleResult:
        """
        Search for the detail of an earthquake by its event id.
        :param event_id: the event id of the earthquake
        :return: SingleResult
        :raises ValueError:  If the HTTP response from the USGS Earthquake API is not 200
        """
        if not isinstance(event_id, str):
            raise TypeError("event id should be a string")
        query_dict = {}
        query_dict["format"] = "geojson"
        query_dict["eventid"] = event_id
        payload_str = urllib.parse.urlencode(query_dict, safe=':')
        url = EarthquakeQuery._base_url + "?" + payload_str
        r = requests.get(url)
        if r.status_code == 200:
            return SingleResult(r.json())
        else:
            raise ValueError(r.text)

    def search(self) -> ResultCollection:
        """
        Search for a collection of results according to the parameters.

        Note:
            The default value for parameter limit is set to 20000, because the USGS earthquake API can support up to
            20000 results in a single query

        :return: ResultCollection, the collection of the results of the query
        """
        result = []
        for time_single in self._query_time:
            for location_single in self._query_location:
                result.append(self._query_single(time_single, location_single))
        result_object = ResultCollection(result)
        return result_object

    def _build_other_extension_params_dic(self) -> dict:
        result = {}

        attributes = self.__dict__
        for key in EarthquakeQuery._attribute_map:
            if EarthquakeQuery._attribute_map[key] in attributes:
                # traverse the _attribute_map to construct the parameters for other and extension
                value = attributes[EarthquakeQuery._attribute_map[key]]
                if value is None:
                    continue
                elif type(value) is Magnitude:
                    result[key] = value.value
                elif type(value) is Origin:
                    result[key] = value.value
                elif type(value) is Delete:
                    result[key] = value.value
                elif type(value) is Supersede:
                    result[key] = value.value
                elif type(value) is Alertlevel:
                    result[key] = value.value
                elif type(value) is Catalog:
                    result[key] = value.value
                elif type(value) is Contributor:
                    result[key] = value.value
                else:
                    result[key] = value
        return result

    def _query_single(self, time: TimeFrame, location: Location):
        # set the format to geojson
        query_dict = {"format": "geojson"}
        # if the time needs to be set
        if time is not None:
            # set start time
            query_dict["starttime"] = time.get_start_time_string()
            # set end time
            query_dict["endtime"] = time.get_end_time_string()
            if time.is_update_after_set():
                # if have update after, set it
                query_dict["updatedafter"] = time.get_update_after_string()
        if location is not None:
            # if the location needs to be set
            query_dict.update(location.get_value())
        # update the query dict with other and extension parameters
        query_dict.update(self._build_other_extension_params_dic())
        # build the parameters in URL
        payload_str = urllib.parse.urlencode(query_dict, safe=':')
        # append the url
        url = EarthquakeQuery._base_url + "?" + payload_str
        r = requests.get(url)
        if r.status_code == 200:
            return r.json()
        else:
            raise ValueError(r.text)

    def get_query_parameters(self) -> dict:
        """
        Get the current parameters in a dict

        :return: Dict, a dict of query parameters that has been set in the query
        """
        query_parameters = {}
        # if the time is not just None
        if self._query_time != [None]:
            # build time list for a single request
            time_list = []
            for one in self._query_time:
                one_dict = {"starttime": one.get_start_time_string(), "endtime": one.get_end_time_string()}
                if one.is_update_after_set():
                    one_dict["updateafter"] = one.get_update_after_string()
                time_list.append(one_dict)
            query_parameters["time"] = time_list
        # if the location is not just None
        if self._query_location != [None]:
            # build location list for a single request
            query_parameters["location"] = [one.get_value() for one in self._query_location]
        attributes = self.__dict__
        for key in EarthquakeQuery._attribute_map:
            if EarthquakeQuery._attribute_map[key] in attributes:
                # get the attribute value
                value = attributes[EarthquakeQuery._attribute_map[key]]
                if value is None:
                    continue
                else:
                    query_parameters[key] = value
        return query_parameters

    def set_time(self, time: List[TimeFrame]) -> 'EarthquakeQuery':
        """
        Set the time ranges for the search query

        :param time: a list of TimeFrame instances
        :return: EarthquakeQuery, self
        """
        if time is None:
            raise TypeError("time shouldn't be none")
        self._query_time = time
        return self

    def get_time(self) -> List[TimeFrame]:
        """
        Get a list of time frames set in the query parameter

        :return: List, the list of time frames
        """
        return self._query_time

    def set_location(self, location: List[Location]) -> 'EarthquakeQuery':
        """
        Set the location of the query

        :param location: a list of Location instances

        :return: EarthquakeQuery, self
        """
        if location is None:
            raise TypeError("location shouldn't be none")
        self._query_location = location
        return self

    def get_location(self) -> List[Location]:
        """
        Get the list of location set in the query parameter

        :return: List, the list of location in the query
        """
        return self._query_location

    def set_catalog(self, catalog: Catalog) -> 'EarthquakeQuery':
        """
        Set the catalog of the earthquake query. Limit the events from a specified catalog.

        Note:
            When catalog and contributor are omitted, the most preferred
            information from any catalog or contributor for the event is returned.

        :param catalog: catalog of the earthquake query
        :type catalog: Catalog
        :return: EarthquakeQuery
        """
        if isinstance(catalog, Catalog):
            self._query_catalog = catalog
            return self
        else:
            raise TypeError("set_catalog input is not with type Catalog")

    def get_catalog(self) -> str:
        """
        Get the value of the catalog parameter

        :return: str, value of the catalog parameter
        """
        return self._query_catalog

    def set_contributor(self, contributor: Contributor) -> 'EarthquakeQuery':
        """
        Set the contributor of the earthquake query.
        Limit to events contributed by a specified contributor.

        Note:
            When catalog and contributor are omitted, the most preferred
            information from any catalog or contributor for the event is returned.

        :param contributor: contributor of the earthquake query.
        :type contributor: Contributor
        :return: EarthquakeQuery, self
        """
        if isinstance(contributor, Contributor):
            self._query_contributor = contributor
            return self
        else:
            raise TypeError("set_contributor input is not with type Contributor")

    def get_contributor(self) -> str:
        """
        Get the value of the contributor parameter

        :return: str, value of the contributor parameter
        """
        return self._query_contributor

    def set_include_all_magnitudes(self, magnitude_flag: Magnitude) -> 'EarthquakeQuery':
        """
        Set whether the query result should include all the magnitude.

        The default magnitude_flag is set to be Magnitude.NOT_INCLUDE_ALL_MAGNITUDE.

        :raise TypeError: If the magnitude_flag is not with type Enum Magnitude
               IllegalFormat: if the format of the query is not xml or quakeml.
        :param magnitude_flag: if magnitude_flag is INCLUDE_ALL_MAGNITUDE, set the query to include all the magnitude;
                               otherwise, set the query not to include all the magnitude
        :type magnitude_flag: Magnitude
        :return: EarthquakeQuery
        """
        if isinstance(magnitude_flag, Magnitude):
            self._query_include_all_magnitudes_flag = magnitude_flag
            return self
        else:
            raise TypeError("set_include_all_magnitudes input is not with type Magnitude")

    def get_include_all_magnitudes(self):
        """
        Get the status of whether the earthquake query result should include all the magnitude.

        :return: status of whether the earthquake query result should include all the magnitude
        """
        return self._query_include_all_magnitudes_flag

    def set_include_all_origins(self, origin_flag: Origin) -> 'EarthquakeQuery':
        """
        Set whether the query result should include all the origins.

        The default origin_flag is set to be Origin.NOT_INCLUDE_ALL_ORIGINS.

        :raise TypeError: If the origin_flag is not with type Enum Origin
                          IllegalFormat: if the format of the query is not xml or quakeml.
        :param origin_flag: if origin_flag is INCLUDE_ALL_ORIGINS, set the query to include all the origins;
                            otherwise, set the query not to include all the origins.
        :type origin_flag: Origin
        :return: EarthquakeQuery
        """
        if isinstance(origin_flag, Origin):
            self._query_include_all_origins_flag = origin_flag
            return self
        else:
            raise TypeError("set_include_all_origins input is not with type Origin")

    def get_include_all_origins(self):
        """
        Get the status of whether the earthquake query result should include all the origins.

        :return: status of whether the earthquake query result should include all the origins
        """
        return self._query_include_all_origins_flag

    def set_include_deleted(self, delete_flag: Delete) -> 'EarthquakeQuery':
        """
        Set whether the query result should include events that are deleted.

        The default delete_flag is set to be Delete.NOT_INCLUDE_DELETED

        :raise  TypeError: If the delete_flag is not with type Enum Delete
                IllegalFormat: if the format of the query is not geojson or csv.
        :param delete_flag: if delete_flag is INCLUDE_DELETED, the query result include the deleted events.
                            otherwise, the query result does not include the deleted events
        :type delete_flag: Delete
        :return: EarthquakeQuery
        """
        if isinstance(delete_flag, Delete):
            self._query_include_all_deleted_flag = delete_flag
            return self
        else:
            raise TypeError("set_include_deleted input is not with type Delete")

    def get_include_deleted(self) -> bool:
        """
        Get the status of whether the earthquake query result should include the deleted events.

        :return: Bool, status of whether the earthquake query result should include the deleted events.
        """
        return self._query_include_all_deleted_flag

    def set_include_superseded(self, superseded_flag: Supersede) -> 'EarthquakeQuery':
        """
        Set the query result to whether include the superseded events or products.

        The default superseded_flag is set to be NOT_INCLUDE_SUPERSEDED.

        Note: This operation will also include all deleted products.

        :raise: TypeError: If the superseded_flag is not with type Enum Supersede
                IllegalState: if the eventid is not specified.
        :param superseded_flag: if supersded_flag is INCLUDE_SUPERSEDED, the query result
                                should include the superseded events of the specified eventid;
                                otherwise, the query result should not included the superseded events.
        :type superseded_flag: Supersede
        :return: EarthquakeQuery
        """
        if isinstance(superseded_flag, Supersede):
            if self._query_eventid is not None:
                self._query_include_all_superseded_flag = superseded_flag
                return self
            else:
                raise ValueError("Query eventid is not specified.")
        else:
            raise TypeError("set_include_superseded input is not with type Supersede")

    def get_include_superseded(self) -> bool:
        """
        Get the status of whether the earthquake query result should include events being superseded.

        :return: Bool, the status of whether the earthquake query result should include events being superseded.
        """
        return self._query_include_all_superseded_flag

    def set_limit(self, limit) -> 'EarthquakeQuery':
        """
        Set the query result to a limit number.

        The limit should be an integer in the rage of 0 (inclusive) and 20000 (inclusive).
        The default limit is set to be 20000.

        :exception IllegalArgument: limit is greater than 20000 or less than 0
                                    limit is not an integer
        :param limit: limit of the earthquake query result
        :type limit: int
        :return: EarthquakeQuery
        """
        if isinstance(limit, int):
            if 0 <= limit <= 20000:
                self._query_limit = limit
            else:
                raise ValueError("set_limit input out of bound")
        else:
            raise TypeError("set_limit input should be an integer")
        return self

    def get_limit(self) -> int:
        """
        Get the limit number of a query result.

        :return: Int, limit number of a query result
        """
        return self._query_limit

    def set_max_depth(self, max_depth) -> 'EarthquakeQuery':
        """
        Set the query with limit to events with depth less than the max_depth.

        The max_depth should be in the rage of -100 (inclusive) and 1000 (inclusive).
        The default value of max_depth is set to be 1000.
        The unit of the depth is km.

        :raise ValueError: max_depth is greater than 1000 or less than -100.
        :param max_depth: max depth of an earthquake query
        :return: EarthquakeQuery
        """

        if max_depth is None:
            raise ValueError("set_max_depth input cannot be None.")
        if isinstance(max_depth, float) or isinstance(max_depth, int):
            if -100 <= max_depth <= 1000:
                self._query_max_depth = max_depth
                return self
            else:
                raise ValueError("set_max_depth input out of bounds")
        else:
            raise ValueError("set_max_depth input should be numeric.")

    def get_max_depth(self) -> float:
        """
        Get the maximum depth parameter set to the query.

        :return: float, maximum depth parameter of the query
        """
        return self._query_max_depth

    def set_max_magnitude(self, max_magnitude) -> 'EarthquakeQuery':
        """
        Set the query with limit to events with magnitude less than the max_magnitude.

        :param max_magnitude: max magnitude of an earthquake query.
        :return: EarthquakeQuery
        """
        if max_magnitude is None:
            raise ValueError("set_max_magnitude input cannot be None.")
        if isinstance(max_magnitude, float) or isinstance(max_magnitude, int):
            self._query_max_magnitude = max_magnitude
            return self
        else:
            raise ValueError("set_max_magnitude input should be numeric.")

    def get_max_magnitude(self) -> float:
        """
        Get the maximum magnitude parameter set to the query.

        :return: float, maximum magnitude parameter of the query
        """
        return self._query_max_magnitude

    def set_min_depth(self, min_depth) -> 'EarthquakeQuery':
        """
        Set the query with limit to events with depth greater than the min_depth.

        The min_depth should be in the rage of -100 (inclusive) and 1000 (inclusive).
        The default value of min_depth is set to be -100.
        The unit of the depth is km.

        :raise ValueError: min_depth is greater than 1000 or less than -100.
        :param min_depth: minimum depth of an earthquake query
        :return: EarthquakeQuery, self
        """
        if min_depth is None:
            raise ValueError("set_min_depth input cannot be None.")
        if isinstance(min_depth, float) or isinstance(min_depth, int):
            if -100 <= min_depth <= 1000:
                self._query_min_depth = min_depth
                return self
            else:
                raise ValueError("set_min_depth input out of bounds")
        else:
            raise ValueError("set_min_depth input should be numeric.")

    def get_min_depth(self) -> float:
        """
        Get the minimum depth of a query.

        :return: float, minimum depth of a query
        """
        return self._query_min_depth

    def set_min_magnitude(self, min_magnitude) -> 'EarthquakeQuery':
        """
        Set the query with limit to events with magnitude greater than the min_magnitude.

        :param min_magnitude: minimum magnitude of an earthquake query.
        :return: EarthquakeQuery
        """
        if min_magnitude is None:
            raise ValueError("set_min_magnitude input cannot be None.")
        if isinstance(min_magnitude, float) or isinstance(min_magnitude, int):
            self._query_min_magnitude = min_magnitude
            return self
        else:
            raise ValueError("set_min_magnitude input should be numeric.")

    def get_min_magnitude(self) -> float:
        """
        Get the minimum magnitude of the earthquake query.

        :return: float, minimum magnitude of the earthquake query
        """
        return self._query_min_magnitude

    def set_alertlevel(self, alert_level: Alertlevel) -> 'EarthquakeQuery':
        """
        Limit to events with a specific PAGER alert level.
        The allowed values are:
            1. alert_level=green: Limit to events with PAGER alert level "green".
            2. alert_level=yellow: Limit to events with PAGER alert level "yellow".
            3. alert_level=orange: Limit to events with PAGER alert level "orange".
            4. alert_level=red: Limit to events with PAGER alert level "red".

        :param alert_level: alert level limit for search results
        :exception IllegalArgument: If the magnitude_flag is not with type Enum Alertlevel
        :return: query itself
        """
        if alert_level is None:
            raise TypeError("IllegalArgument, alert_level input cannot be None")
        if isinstance(alert_level, Alertlevel):
            self._query_alert_level = alert_level
            return self
        raise TypeError("IllegalArgument, alert_level input should be an instance of Alertlevel enum")

    def get_alert_level(self) -> str:
        """
        Get the alert level of an earthquake query.

        :return: Str, alert level of an earthquake query
        """
        return self._query_alert_level

    def set_event_type(self, eventtype: str) -> 'EarthquakeQuery':
        """
        Limit to events of a specific type. NOTE: “earthquake” will filter non-earthquake events.

        :param eventtype: target eventtype string
        :exception: IllegalArgument: If the eventtype is not valid string
        :return: EarthquakeQuery itself
        """
        if eventtype is None:
            raise ValueError("IllegalArgument, eventtype input cannot be None")
        if isinstance(eventtype, str):
            self._query_event_type = eventtype
            return self
        raise ValueError("IllegalArgument, eventtype input should be string type")

    def get_event_type(self) -> str:
        """
        Get the event type of an earthquake query.

        :return: Str, event type of an earthquake query.
        """
        return self._query_event_type

    def set_max_cdi(self, maxcdi: float) -> 'EarthquakeQuery':
        """
        Set maximum cdi limit for query search results.

        :param maxcdi: float, [0,12], Maximum value for Maximum Community Determined Intensity reported by DYFI.
        :exception: IllegalArgument: If the maxcdi is not valid
        :return: EarthquakeQuery itself
        """
        if maxcdi is None:
            raise ValueError("IllegalArgument, maxcdi input cannot be None")
        if not (isinstance(maxcdi, float) or isinstance(maxcdi, int)):
            raise ValueError("IllegalArgument, maxcdi input should be numeric")
        if maxcdi < 0 or maxcdi > 12:
            raise ValueError("IllegalArgument, maxcdi input should be in the range of [0, 12]")
        if self._query_min_cdi is not None and maxcdi <= self._query_min_cdi:
            raise ValueError("IllegalArgument, maxcdi input should be larger than mincdi")
        self._query_max_cdi = maxcdi
        return self

    def get_max_cdi(self) -> float:
        """
        Get the maximum cdi of an earthquake query.

        :return: Flaot, maximum cdi of an earthquake query
        """
        return self._query_max_cdi

    def set_min_cdi(self, mincdi: float) -> 'EarthquakeQuery':
        """
        Set minimum cdi limit for query search results.

        :param mincdi: Decimal, Minimum value for Maximum Community Determined Intensity reported by DYFI.
        :exception IllegalArgument: If the mincdi is not valid
        :return: EarthquakeQuery itself
        """
        if mincdi is None:
            raise ValueError("IllegalArgument, mincdi input cannot be None")
        if not (isinstance(mincdi, float) or isinstance(mincdi, int)):
            raise ValueError("IllegalArgument, mincdi input should be numeric")
        if self._query_max_cdi is not None and self._query_max_cdi <= mincdi:
            raise ValueError("IllegalArgument, mincdi input should be smaller than maxcdi")
        self._query_min_cdi = mincdi
        return self

    def get_min_cdi(self) -> float:
        """
        Get the minimum cdi of an earthquake query.

        :return: Float, minimum cdi of an earthquake query.
        """
        return self._query_min_cdi

    def set_max_gap(self, maxgap: float) -> 'EarthquakeQuery':
        """
        Set the max gap parameter of the query

        :param maxgap: float, [0,360] degrees, Limit to events with no more than this azimuthal gap.
        :return: EarthquakeQuery itself
        """
        if maxgap is None:
            raise ValueError("IllegalArgument, maxgap input cannot be None")
        if not (isinstance(maxgap, float) or isinstance(maxgap, int)):
            raise ValueError("IllegalArgument, maxgap input should be numeric")
        if maxgap < 0 or maxgap > 360:
            raise ValueError("IllegalArgument, maxgap input should be in the range [0,360]")
        if self._query_min_gap is not None and maxgap <= self._query_min_gap:
            raise ValueError("IllegalArgument, maxgap input should be larger than mingap")
        self._query_max_gap = maxgap
        return self

    def get_max_gap(self) -> float:
        """
        Get the maximum gap of an earthquake query.

        :return: float, maximum gap of an earthquake query.
        """
        return self._query_max_gap

    def set_min_gap(self, mingap: float) -> 'EarthquakeQuery':
        """
        Set minimum gap limit for query search results.

        :param mingap: Float, [0,360] degrees, Limit to events with no less than this azimuthal gap.
        :exception IllegalArgument: If the mingap is not valid
        :return: EarthquakeQuery itself
        """
        if mingap is None:
            raise ValueError("IllegalArgument, mingap input cannot be None")
        if not (isinstance(mingap, float) or isinstance(mingap, int)):
            raise ValueError("IllegalArgument, mingap input should be numeric")
        if mingap < 0 or mingap > 360:
            raise ValueError("IllegalArgument, mingap input should be in the range [0,360]")
        if self._query_max_gap is not None and self._query_max_gap <= mingap:
            raise ValueError("IllegalArgument, mingap input should be smaller than maxgap")
        self._query_min_gap = mingap
        return self

    def get_min_gap(self) -> float:
        """
        Get the minimum gap of an earthquake query.

        :return: float, minimum gap of an earthquake query.
        """
        return self._query_min_gap

    def set_max_mmi(self, maxmmi: float) -> 'EarthquakeQuery':
        """
        Set mmi limit for query search results.

        :param maxmmi: Decimal [0,12], Maximum value for Maximum Modified Mercalli Intensity reported by ShakeMap.
        :exception IllegalArgument: If the maxmmi is not valid
        :return: EarthquakeQuery itself
        """
        if maxmmi is None:
            raise ValueError("IllegalArgument, maxmmi cannot be None")
        if not (isinstance(maxmmi, float) or isinstance(maxmmi, int)):
            raise ValueError("IllegalArgument, maxmmi should be numeric")
        if 0 <= maxmmi <= 12:
            self._query_max_mmi = maxmmi
            return self
        raise ValueError("IllegalArgument, maxmmi should be in the range [0,12]")

    def get_max_mmi(self) -> float:
        """
        Get the maximum mmi of an earthquake query.

        :return: Float, maximum mmi of an earthquake query.
        """
        return self._query_max_mmi

    def set_max_significance(self, maxsig: int) -> 'EarthquakeQuery':
        """
        Set maximum significance limit for query search results.

        :param maxsig: Integer, limit to events with no more than this significance.
        :exception IllegalArgument: If the maxsig is not valid
        :return: EarthquakeQuery itself
        """
        if maxsig is None:
            raise ValueError("IllegalArgument, maxsig input cannot be None")
        if not isinstance(maxsig, int):
            raise ValueError("IllegalArgument, maxsig input should be int type")
        if self._query_min_significance is not None and maxsig <= self._query_min_significance:
            raise ValueError("IllegalArgument, maxsig input should be larger than minsig")
        self._query_max_significance = maxsig
        return self

    def get_max_significance(self) -> int:
        """
        Get the maximum significance of an earthquake query.

        :return: int, maximum significance of an earthquake query.
        """
        return self._query_max_significance

    def set_min_significance(self, minsig: int) -> 'EarthquakeQuery':
        """
        Set minimum significance limit for query search results.

        :param minsig: Integer, limit to events with no less than this significance.
        :exception IllegalArgument: If the minsig is not valid
        :return: query itself
        """
        if minsig is None:
            raise ValueError("IllegalArgument, minsig input cannot be None")
        if not isinstance(minsig, int):
            raise ValueError("IllegalArgument, minsig input should be int type")
        if self._query_max_significance is not None and minsig >= self._query_max_significance:
            raise ValueError("IllegalArgument, minsig input should be smaller than maxsig")
        self._query_min_significance = minsig
        return self

    def get_min_significance(self) -> int:
        """
        Get the minimum significance of an earthquake query.

        :return: Int, minimum significance of an earthquake query.
        """
        return self._query_min_significance

    def set_min_feltcount(self, minfelt: int) -> 'EarthquakeQuery':
        """
        Set felt count limit for query search results.

        :param minfelt: Integer[1,∞], limit to events with this many DYFI responses.
        :exception IllegalArgument: If the minfelt is not valid
        :return: EarthquakeQuery itself
        """
        if minfelt is None:
            raise ValueError("IllegalArgument, minfelt input cannot be None")
        if not isinstance(minfelt, int):
            raise ValueError("IllegalArgument, minfelt input should be int type")
        if minfelt >= 1:
            self._query_min_feltcount = minfelt
            return self
        raise ValueError("IllegalArgument, minfelt input should be equal or larger than 1")

    def get_min_feltcount(self) -> int:
        """
        Get the minimum feltcount of an earthquake query.

        :return: Int, minimum feltcount of an earthquake query.
        """
        return self._query_min_feltcount

    def set_nodata_code(self, nodata: int) -> 'EarthquakeQuery':
        """
        Define the error code that will be returned when no data is found.
        :param nodata: Int, (204 or 404)
        :exception IllegalArgument: If the nodata is not valid
        :return query itself
        """
        if nodata is None:
            raise ValueError("IllegalArgument, nodata input cannot be None")
        if not isinstance(nodata, int):
            raise ValueError("IllegalArgument, nodata input should be int type")
        self._query_nodata = nodata
        return self

    def get_nodata_code(self) -> int:
        """
        Get the error code that will be returned when no data is found.

        :return: Int, error code of an earthquake query when no data is found.
        """
        return self._query_nodata

    def set_product_limit(self, producttype: str, productcode: str) -> 'EarthquakeQuery':
        """
        Set product related limit for query search results.

        :param producttype: Limit to events that have this type of product associated
        :param productcode: Limit to the event that is associated with the productcode. The event will be returned even
                            if the productcode is not the preferred code for the event.
        :exception IllegalArgument: If the producttype or productcode is not valid
        :return: EarthquakeQuery itself
        """
        if producttype is None:
            raise ValueError("IllegalArgument, producttype input cannot be None")
        if productcode is None:
            raise ValueError("IllegalArgument, productcode input cannot be None")
        if not isinstance(producttype, str):
            raise ValueError("IllegalArgument, producttype input should be string type")
        if not isinstance(productcode, str):
            raise ValueError("IllegalArgument, productcode input should be string type")
        self._query_product_type = producttype
        self._query_product_code = productcode
        return self

    def get_product_type(self) -> str:
        """
        Get the product type of an earthquake query.

        :return: Str, product type of an earthquake query.
        """
        return self._query_product_type

    def get_product_code(self) -> int:
        """
        Get the product code of an earthquake query.

        :return: int, product code of an earthquake query.
        """
        return self._query_product_code

    def set_reviewstatus(self, reviewstatus: str) -> 'EarthquakeQuery':
        """
        Limit to events with a specific review status.

        :param reviewstatus: reviewstatus=automatic limits to events with review status "automatic". reviewstatus=reviewed:
                             limits to events with review status "reviewed".
        :exception IllegalArgument: If the reviewstatus is not valid
        :return: EarthquakeQuery itself
        """
        if reviewstatus is None:
            raise ValueError("IllegalArgument, reviewstatus input cannot be None")
        if not isinstance(reviewstatus, str):
            raise ValueError("IllegalArgument, reviewstatus input should be string type")
        self._query_review_status = reviewstatus
        return self

    def get_review_status(self) -> str:
        """
        Get the review status of an earthquake query.

        :return: Str, review status of an earthquake query.
        """
        return self._query_review_status



from enum import Enum


class Magnitude(Enum):
    """
    Enum values of includeallmagnitudes flag, which indicates whether all
    magnitudes for the event should be included.
    """
    INCLUDE_ALL_MAGNITUDE = True
    """
    Include all magnitude in the query result
    """

    NOT_INCLUDE_ALL_MAGNITUDE = False
    """
    Not include all magnitude in the query result
    """



class _Address:
    """
    Private class for representing addresses. Not part of the API.
    """
    def __init__(self):
        self.address_line = None
        self.admin_district = None
        self.admin_district2 = None
        self.country_region = None
        self.country_region_iso2 = None
        self.formatted_address = None
        self.locality = None
        self.postal_code = None

    def set_address_line(self, address_line) -> '_Address':
        self.address_line = address_line
        return self

    def set_admin_district(self, admin_district) -> '_Address':
        self.admin_district = admin_district
        return self

    def set_admin_district2(self, admin_district2) -> '_Address':
        self.admin_district2 = admin_district2
        return self

    def set_country_region(self, country_region) -> '_Address':
        self.country_region = country_region
        return self

    def set_country_region_iso2(self, country_region_iso2) -> '_Address':
        self.country_region_iso2 = country_region_iso2
        return self

    def set_formatted_address(self, formatted_address) -> '_Address':
        self.formatted_address = formatted_address
        return self

    def set_locality(self, locality) -> '_Address':
        self.locality = locality
        return self

    def set_postal_code(self, postal_code) -> '_Address':
        self.postal_code = postal_code
        return self

    def get_address_line(self):
        return self.address_line

    def get_admin_district(self):
        return self.admin_district

    def get_admin_district2(self):
        return self.admin_district2

    def get_country_region(self):
        return self.country_region

    def get_formatted_address(self):
        return self.formatted_address

    def get_locality(self):
        return self.locality

    def get_postal_code(self):
        return self.postal_code

    def get_value(self):
        return self.__dict__


from enum import Enum

class Nodata(Enum):
    """
    Enum representing the values of the error code when no data is found, in the query parameters.
    """
    NO_CONTENT = 204
    """
    204 - No content
    """

    NOT_FOUND = 404
    """
    404 - Not found
    """



from enum import Enum


class Delete(Enum):
    """
    Enum that represents the includedeleted flag in the query parameters
    """
    INCLUDE_DELETED = True
    """
    Include deleted events in the query result
    """

    NOT_INCLUDE_DELETED = False
    """
    Not include deleted events in the query result
    """

    ONLY_DELETED = "only"
    """
    Only include deleted events in the query result
    """



from abc import ABC, abstractmethod
from enum import Enum
from .geocode import _GeocodeQuery
from .key import _Key


class Location(ABC):
    """
    This is the location abstract class. After initiating the locations, clients can build a list
    of the Location object, and put it into the EarthquakeQuery constructor.
    There are 4 types of locations: Rectangle, Circle, GeoRectangle and GeoCircle. GeoRectangle and
    GeoCircle use Bing Map's "Find a Location by Query" API to search the coordinates of the location.

    Example:
    ::
        rectangle = Rectangle(36.458534, -123.399768, 38.571488, -120.927844)
        circle = Circle(36.458534, -123.399768, RadiusUnit.DEGREE, 30)
        geo_rectangle = GeoRectangle("Pittsburgh, PA")
        geo_circle = GeoCircle("San Francisco", RadiusUnit.KM, 40)
        earthquake_query = EarthquakeQuery(time=None, Location=[rectangle, circle, geo_rectangle, geo_circle])
        result = earthquake_query.search()

    To show the information about the location, use the method show.
    Example:
    ::
        rectangle1 = Rectangle(0, 0, 90, 90)
        rectangle1.show()
    The console would print:
        Rectangle with min_latitude 0, min_longitude 0, max_latitude 90, max_longitude 90.
    """

    @abstractmethod
    def get_type(self) -> str:
        """
        Abstract method for getting the type of the location
        :return: dict, the type of the location
        """
        pass

    def get_value(self) -> dict:
        """
        Abstract method for getting the value dict of the location
        :return: dict, the values of the location
        """
        pass

    @abstractmethod
    def show(self):
        """
        Abstract method for printing the information of the location
        """
        pass


class Rectangle(Location):
    def __init__(self, min_latitude=-90, min_longitude=-180, max_latitude=90, max_longitude=180):
        """
        Build a rectangle location. Requests may use any combination of these parameters.
        :param min_latitude: Limit to events with a latitude larger than the specified minimum.
                             Default value is -90.0. The range is [-90.0, 90.0] degrees.
                             NOTE: min_latitude must be less than max_latitude.
        :type min_latitude:  float
        :param min_longitude: Limit to events with a longitude larger than the specified minimum.
                             Default value is -180.0. The range is [-360.0, 360.0] degrees.
                             NOTE: rectangles may cross the date line by using a min_longitude < -180 or
                                   max_longitude > 180.
                             NOTE: min_longitude must be less than max_longitude.
        :type min_longitude:  float
        :param max_latitude: Limit to events with a latitude smaller than the specified maximum.
                             Default value is 90. The range is [-90.0, 90.0] degrees.
                             NOTE: min_latitude must be less than max_latitude.
        :type max_latitude:  float
        :param max_longitude: Limit to events with a longitude smaller than the specified maximum.
                             Default value is 180.0. The range is [-360.0, 360.0] degrees.
                             NOTE: rectangles may cross the date line by using a min_longitude < -180 or
                                   max_longitude > 180.
                             NOTE: min_longitude must be less than max_longitude.
        :type max_longitude:  float
        :raises TypeError:  If any parameter is not type float or cannot convert to float.
        :raises ValueError: If min_latitude, min_longitude, max_latitude, max_longitude is not within the range.
        :raises ValueError: If min_longitude is not less than max_longitude or
                            min_latitude is not less than max_latitude.
        :return: None
        """
        try:
            float(min_latitude)
        except:
            raise TypeError("Cannot convert min_latitude to float")

        try:
            float(min_longitude)
        except:
            raise TypeError("Cannot convert min_longitude to float")

        try:
            float(max_latitude)
        except:
            raise TypeError("Cannot convert max_latitude to float")

        try:
            float(max_longitude)
        except:
            raise TypeError("Cannot convert max_longitude to float")

        if not -90.0 <= min_latitude <= 90.0:
            raise ValueError("The range of min_latitude is [-90.0, 90.0] degrees.")

        if not -360.0 <= min_longitude <= 360.0:
            raise ValueError("The range of min_longitude is [-360.0, 360.0] degrees.")

        if not -90.0 <= max_latitude <= 90.0:
            raise ValueError("The range of max_latitude is [-90.0, 90.0] degrees.")

        if not -360.0 <= max_longitude <= 360.0:
            raise ValueError("The range of max_longitude is [-360.0, 360.0] degrees.")

        if not min_latitude < max_latitude:
            raise ValueError("min_latitude should be smaller than max_latitude")

        if not min_longitude < max_longitude:
            raise ValueError("min_longitude should be smaller than max_longitude")

        self.min_latitude = min_latitude
        self.min_longitude = min_longitude
        self.max_latitude = max_latitude
        self.max_longitude = max_longitude
        self.type = "rectangle"

    def get_min_latitude(self) -> float:
        """
        Get the min latitude.
        :return: Float, the value of min_latitude
        """
        return self.min_latitude

    def get_max_latitude(self) -> float:
        """
        Get the max latitude.
        :return: Float, the value of max_latitude
        """
        return self.max_latitude

    def get_min_longitude(self) -> float:
        """
        Get the min longitude.
        :return: Float, the value of min_longitude
        """
        return self.min_longitude

    def get_max_longitude(self) -> float:
        """
        Get the max longitude.
        :return: Float, the value of max_longitude
        """
        return self.max_longitude

    def show(self):
        """
        Print the information of the location.
        """
        print("[Rectangle] min_latitude: " + str(self.min_latitude) +
              ", min_longitude: " + str(self.min_longitude) +
              ", max_latitude: " + str(self.max_latitude) +
              ", max_longitude: " + str(self.max_longitude))

    def get_type(self) -> str:
        """
        Get the type of the Location.
        :return: The type of the location.
        """
        return self.type

    def get_value(self) -> dict:
        """
        Get the value dictionary, used in EarthquakeQuery class.
        :return: The value dict.
        """
        result = {}
        result["minlatitude"] = self.min_latitude
        result["minlongitude"] = self.min_longitude
        result["maxlatitude"] = self.max_latitude
        result["maxlongitude"] = self.max_longitude
        return result


class RadiusUnit(Enum):
    """
    The Circle object's radius unit.
    """
    KM = 0, 'kilometer'
    """
    Kilometer, default is 0
    """

    DEGREE = 1, 'a degree of latitude/longitude'
    """
    Degree, 1 degree equals 1 degree of latitude and longitude, default is 1
    """


class Circle(Location):
    def __init__(self, latitude, longitude, radius_unit=RadiusUnit.KM, radius=180):
        """
        Build a circle location. We support 2 radius units, km and degree.
        Example:
        ::
            circle1 = Circle(36.458534, -123.399768, RadiusUnit.DEGREE, 30)
            circle2 = Circle(36.458534, -123.399768, RadiusUnit.KM, 30)

        :param latitude: Specify the latitude to be used for a radius search. The range is [-90.0, 90.0] degrees.
        :type latitude: float
        :param longitude: Specify the longitude to be used for a radius search. The range is [-180.0, 180.0] degrees.
        :type longitude: float
        :param radius: Limit to events within the specified maximum number of kilometers from the geographic
                       point defined by the latitude and longitude parameters.
                       The range of radius km is [0, 20001.6] km.
                       The range of radius degree is [0, 180.0] degrees.
        :type radius: float
        :param radius_unit: The unit of radius.
        :type radius_unit: RadiusUnit
        :raises TypeError:  If latitude, longitude or radius is not type float or cannot convert to float.
        :raises ValueError: If radius_unit is not type of RadiusUnit.
        :raises ValueError: If latitude, longitude or radius is not within the range.
        """
        try:
            float(latitude)
        except:
            raise TypeError("Cannot convert latitude to float")

        try:
            float(longitude)
        except:
            raise TypeError("Cannot convert longitude to float")

        if not -90.0 <= latitude <= 90.0:
            raise ValueError("The range of latitude is [-90.0, 90.0] degrees.")

        if not -360.0 <= longitude <= 360.0:
            raise ValueError("The range of longitude is [-360.0, 360.0] degrees.")

        if not isinstance(radius_unit, RadiusUnit):
            raise TypeError("radius_unit should be instance of RadiusUnit.")

        try:
            float(radius)
        except:
            raise TypeError("Cannot convert radius to float")

        if radius_unit == RadiusUnit.KM:
            if not 0 <= radius <= 20001.6:
                raise ValueError("The range of radius is [0, 20001.6] km.")

        elif radius_unit == RadiusUnit.DEGREE:
            if not 0 <= radius <= 180:
                raise ValueError("The range of max_radius_degree is [0, 180] degrees.")
        else:
            raise ValueError("No such radius")

        self.type = "circle"
        self.latitude = latitude
        self.longitude = longitude
        self.radius_unit = radius_unit
        self.radius = radius

    def get_longitude(self) -> float:
        """
        Get the value of longitude.
        :return: Float, the value of longitude.
        """
        return self.longitude

    def get_latitude(self) -> float:
        """
        Get the value of latitude.
        :return: Float, the value of latitude.
        """
        return self.latitude

    def get_radius(self) -> float:
        """
        Get the value of radius.
        :return: Float, the value of radius.
        """
        return self.radius

    def get_radius_unit(self) -> float:
        """
        Get the unit if radius.
        :return: Float, the radius unit.
        """
        return self.radius_unit

    def show(self):
        """
        Print the information of the location.
        """
        print("[Circle] latitude: " + str(self.latitude) +
              ", longitude: " + str(self.longitude) +
              ", radius: " + str(self.radius) +
              ", unit: " + str(self.radius_unit))

    def get_type(self) -> str:
        """
        Get the type of the Location.
        :return: str, the type of the location.
        """
        return self.type

    def get_value(self) -> dict:
        """
        Get the value dictionary, used in EarthquakeQuery class.
        :return: Dict, the value dict.
        """
        result = {}
        result["latitude"] = self.latitude
        result["longitude"] = self.longitude
        if self.radius_unit == RadiusUnit.KM:
            result["maxradiuskm"] = self.radius
        else:
            result["maxradius"] = self.radius
        return result


class GeoRectangle(Rectangle):
    def __init__(self, address, latitude_distance_degree=None, longitude_distance_degree=None):
        """
        The GeoCircle class use the Bing Maps "Find a Location by Query" API to search coordinates of
        the address and initialize a rectangle shape location by the search result.
        If user provide latitude_distance_degree and longitude_distance degree, we will use the center of the location,
        and initialize a rectangle by it.
        If user do not specify the desired latitude/longitude, we will use the bounding box provided by Bing Maps.

        Example:
        ::
            geo_rectangle1 = GeoRectangle("Pittsburgh, PA")
            geo_rectangle2 = GeoRectangle("Carnegie Mellon University")
            geo_rectangle3 = GeoRectangle("5000 Forbes Ave, Pittsburgh, PA 15213, United States", 30, 30)

        To use this GeoLocation functionality, please go to
        https://docs.microsoft.com/en-us/bingmaps/getting-started/bing-maps-dev-center-help/getting-a-bing-maps-key
        and get a bing maps API key, and put it in the ./src/key.txt

        :param address: The address of the location.
        :type address: str
        :param latitude_distance_degree: The distance between min_latitude and max_latitude.
                                         Range:(0, 180.0] (Should be larger than zero.), Unit: degree.
        :type latitude_distance_degree: float
        :param longitude_distance_degree: The distance of min_longitude and max_longitude.
                                         Range: (0, 360.0] (Should be larger than zero.), unit: degree.
        :type latitude_distance_degree: float
        :raises TypeError: If the address is not string.
        :raises TypeError: If latitude_distance_degree or longitude_distance_degree is not None and is not type of
                           float or cannot convert to float.
        :raises ValueError: If latitude_distance_degree or longitude_distance_degree is out of range.
        :raises ValueError: If the API cannot find an location for the provided address.
        """

        if not isinstance(address, str):
            raise TypeError("Address is not type str.")

        if _Key.key_path is None:
            raise ValueError("To use geocode, you need to specify the key path first!")

        if latitude_distance_degree is not None and longitude_distance_degree is not None:
            try:
                float(latitude_distance_degree)
            except:
                raise TypeError("Cannot convert latitude_distance_degree to float")

            try:
                float(longitude_distance_degree)
            except:
                raise TypeError("Cannot convert longitude_distance_degree to float")

            if not 0 < latitude_distance_degree <= 180:
                raise ValueError("The range of latitude_distance_degree is [0, 180.0] degrees.")

            if not 0 < longitude_distance_degree <= 360:
                raise ValueError("The range of longitude_distance_degree is [0, 360.0] degrees.")

        geocode_query = _GeocodeQuery(_Key.key_path)
        geocode_query.set_one_line_Address(address)
        json_response = geocode_query.send_query()
        address_location = geocode_query.get_address_boxing(json_response)

        if address_location is None:
            raise ValueError(
                "We are not able to find the location of the address. Please make the address more specific.")

        if latitude_distance_degree is not None and longitude_distance_degree is not None:
            center = address_location.get_center()
            center_latitude, center_longitude = center

            min_latitude = max(-90, center_latitude - latitude_distance_degree / 2.0)
            min_longitude = max(-360, center_longitude - longitude_distance_degree / 2.0)
            max_latitude = min(90, center_latitude + latitude_distance_degree / 2.0)
            max_longitude = min(360, center_longitude + longitude_distance_degree / 2.0)
        else:
            min_latitude, min_longitude, max_latitude, max_longitude = address_location.get_bounding_box()

        super().__init__(min_latitude, min_longitude, max_latitude, max_longitude)
        self.address = address_location.get_address().get_formatted_address()

    def get_address(self) -> str:
        """
        Get the formatted address of the location. The formatted address is the official address of the location,
        for example, if the address "Pittsburgh" is searched, the formatted address would be "Pittsburgh, PA".

        :return: str, the formatted address.
        """
        return self.address

    def show(self):
        """
        Print the information of the location.
        """
        print("[GeoLocation] min_latitude: " + str(self.min_latitude) +
              ", min_longitude: " + str(self.min_longitude) +
              ", max_latitude: " + str(self.max_latitude) +
              ", max_longitude: " + str(self.max_longitude) +
              ", address: " + self.address)


class GeoCircle(Circle):
    def __init__(self, address, radius_unit=RadiusUnit.KM, radius=180):
        """
        The GeoCircle class use the Bing Maps "Find a Location by Query" API to search latitude and longitude of
        the address and initialize a circle shape location by the search result.

        Example:
        ::
            geo_circle1 = GeoCircle("Pittsburgh, PA", radius_unit=RadiusUnit.KM, radius=180)
            geo_circle2 = GeoCircle("Carnegie Mellon University", radius_unit=RadiusUnit.degree, radius=10)

        To use this GeoCircle functionality, please go to
        https://docs.microsoft.com/en-us/bingmaps/getting-started/bing-maps-dev-center-help/getting-a-bing-maps-key
        and get a bing maps API key, and put it in the ./src/key.txt

        :param address: The address of the location.
        :type address: str
        :param radius: Limit to events within the specified maximum number of kilometers from the geographic
                       point defined by the latitude and longitude parameters.
                       The range of radius km is [0, 20001.6] km.
                       The range of radius degree is [0, 180.0] degrees.
        :type radius: float
        :param radius_unit: The unit of radius.
        :type radius_unit: RadiusUnit
        :raises TypeError: If the address is not string.
        :raises ValueError: If the API cannot find an location for the provided address.
        :raises ValueError: If radius is out of range.
        """
        if not isinstance(address, str):
            raise TypeError("Address is not type str.")

        if _Key.key_path is None:
            raise ValueError("To use geocode, you need to specify the key path first!")

        geocode_query = _GeocodeQuery(_Key.key_path)
        geocode_query.set_one_line_Address(address)
        json_response = geocode_query.send_query()
        address_location = geocode_query.get_address_boxing(json_response)

        if address_location is None:
            raise ValueError(
                "We are not able to find the location of the address. Please make the address more specific.")

        center = address_location.get_center()
        super().__init__(center[0], center[1], radius_unit, radius)
        self.address = address_location.get_address().get_formatted_address()

    def get_address(self) -> str:
        """
        Get the formatted address of the location. The formatted address is the official address of the location,
        for example, if the address "Pittsburgh" is searched, the formatted address would be "Pittsburgh, PA".

        :return: str, the formatted address.
        """
        return self.address

    def show(self):
        """
        Print the information of the location.
        """
        print("[GeoCircle] latitude: " + str(self.latitude) +
              ", longitude: " + str(self.longitude) +
              ", radius: " + str(self.radius) +
              ", unit: " + str(self.radius_unit) +
              ", address: " + self.address)



from enum import Enum


class Catalog(Enum):
    """
    A Enum that represents all earthquake data Catalogs. Use this enum in EarthquakeQuery class to set its parameters
    related to catalogs.
    """
    CATALOG_AACSE = "aacse"
    """
    AACSE - Alaska Amphibious Community Seismic Experiment 
    """

    CATALOG_AK = "ak"
    """
    AK - Alaska Earthquake Center
    """

    CATALOG_AT = "at"
    """
    AT - National Tsunami Warning Center
    """

    CATALOG_ATLAS = "atlas"
    """
    ATLAS - ShakeMap Atlas
    """

    CATALOG_AV = "av"
    """
    AV - Alaska Volcano Observatory
    """

    CATALOG_CDMG = "cdmg"
    """
    CDMG - California Division of Mines and Geology
    """

    CATALOG_CGS = "cgs"
    """
    California Geological Survey
    """

    CATALOG_CHOY = "choy"
    """
    CHOY - Energy Magnitude and Broadband Depth
    """

    CATALOG_CI = "ci"
    """
    CI - California Integrated Seismic Network: Southern, California Seismic Network (Caltech/USGS Pasadena and Partners) 
    and Southern California Earthquake Data Center
    """

    CATALOG_CIDEV = "cidev"
    """
    CIDEV - CIDEV database
    """

    CATALOG_DR = "dr"
    """
    DR = DR database
    """

    CATALOG_DUPUTEL = "duputel"
    """
    DUPUTEL - Duputel et al. W phase catalog
    """

    CATALOG_EQH = "eqh"
    """
    EQH - Coffman, von Hake and Stover, Earthquake History of the United States
    """

    CATALOG_EW = "ew"
    """
    EW - EW database
    """

    CATALOG_EW_DM = "ew_dm"
    """
    EW_DM - EW_DM database
    """

    CATALOG_GCMT = "gcmt"
    """
    GCMT - Lamont-Doherty Earth Observatory Global CMT project, New York, USA
    """

    CATALOG_GSC = "gsc"
    """
    Geological Survey of Canada
    """

    CATALOG_HV = "hv"
    """
    HV - Hawaii Volcano Observatory
    """

    CATALOG_ID = "id"
    """
    ID - ID database
    """

    CATALOG_IS = "is"
    """
    IS - IS database
    """

    CATALOG_ISCGEM = "iscgem"
    """
    ISCGEM - ISC-GEM Main Catalog
    """

    CATALOG_ISCGEMSUP = "iscgemsup"
    """
    ISCGEMSUP - ISC-GEM Supplementary Catalog
    """

    CATALOG_ISMPKANSAS = "ismpkansas"
    """
    ISMPKANSAS - USGS Induced Seismicity Project (Kansas)
    """

    CATALOG_LD = "ld"
    """
    LD - Lamont-Doherty Cooperative Seismographic Network
    """

    CATALOG_MB = "mb"
    """
    MB - Montana Bureau of Mines and Geology
    """

    CATALOG_NC = "nc"
    """
    NC - California Integrated Seismic Network: Northern California Seismic System 
    (UC Berkeley, USGS Menlo Park, and Partners)
    """

    CATALOG_NE = "ne"
    """
    NE - New England Seismic Network
    """

    CATALOG_NM = "nm"
    """
    NM - New Madrid Seismic Network
    """

    CATALOG_NN = "nn"
    """
    NN - Nevada Seismological Laboratory
    """

    CATALOG_OFFICIAL = "official"
    """
    OFFICIAL - USGS Earthquake Magnitude Working Group
    """

    CATALOG_OFFICIAL19631013051759_30 = "official19631013051759_30"
    """
    OFFICIAL - USGS Earthquake Magnitude Working Group
    """

    CATALOG_OK = "ok"
    """
    OK - Oklahoma Geological Survey Statewide Network, Norman, USA (aka TUL, OGS)
    """

    CATALOG_OTT = "ott"
    """
    Geological Survey of Canada, Ottawa, Canada
    """

    CATALOG_PR = "pr"
    """
    PR - Puerto Rico Seismic Network
    """

    CATALOG_PT = "pt"
    """
    PT - Pacific Tsunami Warning Center
    """

    CATALOG_SC = "sc"
    """
    SC - SC database
    """

    CATALOG_SE = "se"
    """
    SE - Center for Earthquake Research and Information
    """

    CATALOG_TX = "tx"
    """
    TX - Texas Seismological Network (TexNet) (aka TEXNET, BEG)
    """

    CATALOG_UNKNOWN = "unknown"
    """
    UNKNOWN = Unknown database
    """

    CATALOG_US = "us"
    """
    US - USGS National Earthquake Information Center, PDE
    """

    CATALOG_USAUTO = "usauto"
    """
    USAUTO - USAUTO database
    """

    CATALOG_USHIS = "ushis"
    """
    USHIS - Stover and Coffman, Seismicity of the United States, 1568-1989
    """

    CATALOG_UU = "uu"
    """
    UU - University of Utah Seismograph Stations
    """

    CATALOG_UW = "uw"
    """
    UW - Pacific Northwest Seismic Network
    """

    CATALOG_EQUALC = "=c"
    """
    =C - =C Database
    """

    CATALOG_38457511 = "38457511"
    """
    38457511 - 38457511 Database
    """


from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pyquakes',
    version='0.0.1',
    author='Jianan Shen, Chen Mo',
    author_email="shenjianan97@gmail.com, mochen1228@gmail.com",
    license_files=('LICENSE.txt',),
    description=('PyQuakes is a Python wrapper for USGS Earthquake Catalog API that manages data querying and '
                 'result processing.'),
    package_dir={'pyquakes': 'src'},
    packages=['pyquakes', 'pyquakes.enum'],
    install_requires=[
        "requests >= 2.15.0"
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/shenjianan97/PyQuakes'
)
